(***********************************************************************)
(*                                                                     *)
(*      moniteur:  lancement des processus                             *)
(*                 attente du lancement                                *)
(*                 demarrage du calcul                                 *)
(*                                                                     *)
(*     nocf  proto en  Camlp4                                          *)
(*                                                                     *)
(*     Emmanuel Chailloux, Equipe PPS, universite P7, projet CARAML    *)
(*                                                                     *)
(*     $Id: mon.ml,v 1.1 2003/06/01 20:48:03 emmanuel Exp $         *)
(*                                                                     *)
(***********************************************************************)

let debug = ref true;;

let usage () = 
  Printf.printf "Usage: mon filename port depth machine_1 ... machine_n\n";
  flush stdout;;

let filename = ref "";;
let port = ref 0;;
let depth = ref 0;;
let machines = ref [||];;
let ip_addrs = ref [||];;

let machines_str = ref "";;
let nodes = ref 0;;
let num_ready = ref 0;;

let m = Mutex.create();;
let c = Condition.create();;

let decode_args  () = 
  let l = Array.length (Sys.argv) in 
  if l < 5 then (usage();  exit 1)
  else 
    nodes := l -4;
    filename := Sys.argv.(1);
    port := int_of_string (Sys.argv.(2));
    depth := int_of_string (Sys.argv.(3));
    machines := Array.create (!nodes) (Sys.argv.(4));
    for i=0 to !nodes-1 do !machines.(i) <- Sys.argv.(i+4) done;
    machines_str := Array.fold_right (fun x y -> x ^ " " ^ y)  !machines "";
    ip_addrs := 
      Array.init !nodes
       (fun i -> (Unix.gethostbyname (!machines.(i))).Unix.h_addr_list.(0))

;;

let fun_service sd sa = 
  if !debug then ( Printf.printf "starting fun_service\n"; flush stdout);
  let line = Utils.my_input_line sd in 
    match (Marshal.from_string line 0) with
      "START",i -> if !debug then (Printf.printf "fun_service : START %d\n" i; flush stdout);
                   Mutex.lock m; incr num_ready; 
                   Condition.broadcast c; Mutex.unlock m
    | "END",i -> if !debug then (Printf.printf "fun_service : END %d\n" i; flush stdout); 
                   Mutex.lock m; decr num_ready; 
                   Condition.broadcast c; Mutex.unlock m
    |  str,i ->  Printf.printf "[] %s\n" str; flush stdout
;;
 
let launch_process () = 
 for i=0 to !nodes-1 do
   let com = "ssh "^ (!machines.(i)) ^ " " ^ !filename ^ " " ^ 
              (string_of_int i) ^ " " ^ 
              (string_of_int !nodes) ^ " " ^ (string_of_int (!port+i+1)) ^ " " ^
              (string_of_int !depth) ^ " "^ (Unix.gethostname() ) ^" " ^
              (string_of_int !port) ^ " " ^ !machines_str 
             ^ " & "
   in 
     ignore(Sys.command ("ssh "^(!machines.(i))^ "  ls -lst /tmp &"));
     ignore(Sys.command com);
     Printf.printf ("[%d] : Sent : |%s|\n") i com ; flush stdout
done;;

let send_addrs num i p = 
  let (ic,oc) = ThreadUnix.open_connection (Unix.ADDR_INET(i,p)) 
  in
    output_string oc ((Marshal.to_string (num, 0, 0) [])^"\n");
    flush oc;
    Unix.shutdown_connection ic;;

let wait_begin () = 
  while (!nodes <> !num_ready) do Condition.wait c m done;
  Printf.printf "Everybody is ready!!\n"; flush stdout
;;

let wait_end () = 
  while (!num_ready <> 0) do Condition.wait c m done;
  Printf.printf "Everybody is finished!!\n"; flush stdout
;;

let start_process () = 
  for i = 0 to (!nodes - 1) do
    send_addrs !nodes  !ip_addrs.(i) (!port+i+1)
  done;;

let main() = 
  decode_args ();
  ignore(Utils.start_server fun_service (Utils.get_my_addr()) !port  !nodes);
  launch_process();
  wait_begin();
  start_process();
  wait_end();;

main();;
 
