(***********************************************************************)
(*                                                                     *)
(*      parser ocaml  flight,  proto en  Camlp4                        *)
(*                                                                     *)
(*     Bruno Baia & Clement Garrigou :                                 *)
(*                     TER maitrise Paris 6 printemps 2001             *)
(*     modifie par                                                     *)
(*       Emmanuel Chailloux :                                          *)
(*                     Equipe PPS, universite P7, projet CARAML        *)
(*                                                                     *)
(*     $Id: pa_ocf.ml,v 1.3 2003/06/01 20:48:03 emmanuel Exp $                                                            *)
(*                                                                     *)
(***********************************************************************)

open Stdpp;;
open Pcaml;;
open Pa_o;;

(* fichier pa_ocf.ml, syntaxe de OCaml Flight *)

(* Synchronized expression *)

let  num_get = ref 0
;;

let num_sync = ref 0
;;

let env = ref []
;;

let add_expr loc body =  
  match !env with
   [] -> body   
  | c -> ( env := []; 
      let e1 = <:expr< do { $list:c$}>> in 
      <:expr<  do { 
                  Flight.reset_env_buffer ();
                  $e1$; $body$ } >>)
;;

let  sync_expr = Grammar.Entry.create gram "expression"
;;

let init = ref false;;

Grammar.Unsafe.clear_entry implem;;
Pcaml.parse_implem := Grammar.Entry.parse implem;;

EXTEND
  GLOBAL: implem sync_expr expr;

implem:
[ [ si = str_item_semi; (sil, stopped) = SELF -> ((si :: sil), stopped)
  | "#"; n = LIDENT; dp = OPT expr; ";;" ->
          ([(<:str_item< # $n$ $opt:dp$ >>, loc)], true)
  | EOI -> ([<:str_item<Flight.num_sync.val := 
            Flight.num_sync.val + $int:string_of_int !num_sync$>>,loc], false) 
]];

  str_item_semi:
    [ [ si = str_item; OPT ";;" -> if !init then  (si, loc)  else
          let op = <:str_item<value ___local_shift = Flight.num_sync.val>> in 
          init := true;
          let l = op :: si :: [] in 
          (<:str_item<declare $list:l$ end>>, loc)
    ] ]
  ;

  (* Core expressions *)
  sync_expr:
    [ "top" RIGHTA
      [  e1 = SELF; ";"; e2 = SELF ->
           <:expr<  do { $list:(e1 :: (get_seq e2))$ }  >>
      |  e1 = SELF; ";" -> e1 ]

    | "sync_expr1"
      [ "get"; e = sync_expr; "from";  n = sync_expr ->
	   begin 
             incr num_get;
             let ne = <:expr< Flight.save_closure 
               (Marshal.to_string (fun() -> $e$) [Marshal.Closures]) >> 
             in 
	       env := !env @ [ne];
	        <:expr< if $n$ == Flight.local then $e$ else
	          Flight.drequest $n$ $int:string_of_int !num_get$ 
                >>
            end

     | "if"; e1 = sync_expr; "then"; e2 = sync_expr LEVEL "sync_expr1";
          e3 = [ "else"; e = sync_expr LEVEL "sync_expr1" -> e 
	       | -> <:expr< () >> ] ->
	    <:expr< if $e1$ then $e2$ else $e3$ >> ]  

    | [ e = sync_expr; ","; el = LIST1 NEXT  SEP "," ->
          <:expr< ( $list:(e :: el)$ ) >> ]

    | "||" RIGHTA
      [ e1 = SELF; "or"; e2 = SELF -> <:expr< $lid:"or"$ $e1$ $e2$ >>
      | e1 = SELF; "||"; e2 = SELF -> <:expr< $e1$ || $e2$ >> ]
    | "&&" RIGHTA
      [ e1 = SELF; "&"; e2 = SELF -> <:expr< $lid:"&"$ $e1$ $e2$ >>
      | e1 = SELF; "&&"; e2 = SELF -> <:expr< $e1$ && $e2$ >> ]
    | "<" LEFTA
      [ e1 = SELF; "<"; e2 = SELF -> <:expr< $e1$ < $e2$ >>
      | e1 = SELF; ">"; e2 = SELF -> <:expr< $e1$ > $e2$ >>
      | e1 = SELF; "<="; e2 = SELF -> <:expr< $e1$ <= $e2$ >>
      | e1 = SELF; ">="; e2 = SELF -> <:expr< $e1$ >= $e2$ >>
      | e1 = SELF; "="; e2 = SELF -> <:expr< $e1$ = $e2$ >>
      | e1 = SELF; "<>"; e2 = SELF -> <:expr< $e1$ <> $e2$ >>
      | e1 = SELF; "=="; e2 = SELF -> <:expr< $e1$ == $e2$ >>
      | e1 = SELF; "!="; e2 = SELF -> <:expr< $e1$ != $e2$ >>
      | e1 = SELF; "$"; e2 = SELF -> <:expr< $lid:"\$"$ $e1$ $e2$ >>
      | e1 = SELF; op = infixop0; e2 = SELF -> <:expr< $lid:op$ $e1$ $e2$ >> ]
    | "^" RIGHTA
      [ e1 = SELF; "^"; e2 = SELF -> <:expr< $e1$ ^ $e2$ >>
      | e1 = SELF; "@"; e2 = SELF -> <:expr< $e1$ @ $e2$ >>
      | e1 = SELF; op = infixop1; e2 = SELF -> <:expr< $lid:op$ $e1$ $e2$ >> ]
    | RIGHTA
      [ e1 = SELF; "::"; e2 = SELF -> <:expr< [$e1$ :: $e2$] >> ]
    | "+" LEFTA
      [ e1 = SELF; "+"; e2 = SELF -> <:expr< $e1$ + $e2$ >>
      | e1 = SELF; "-"; e2 = SELF -> <:expr< $e1$ - $e2$ >>
      | e1 = SELF; "+."; e2 = SELF -> <:expr< $e1$ +. $e2$ >>
      | e1 = SELF; "-."; e2 = SELF -> <:expr< $e1$ -. $e2$ >>
      | e1 = SELF; op = infixop2; e2 = SELF -> <:expr< $lid:op$ $e1$ $e2$ >> ]
    | "*" LEFTA
      [ e1 = SELF; "*"; e2 = SELF -> <:expr< $e1$ * $e2$ >>
      | e1 = SELF; "/"; e2 = SELF -> <:expr< $e1$ / $e2$ >>
      | e1 = SELF; "*."; e2 = SELF -> <:expr< $e1$ *. $e2$ >>
      | e1 = SELF; "/."; e2 = SELF -> <:expr< $e1$ /. $e2$ >>
      | e1 = SELF; "land"; e2 = SELF -> <:expr< $e1$ land $e2$ >>
      | e1 = SELF; "lor"; e2 = SELF -> <:expr< $e1$ lor $e2$ >>
      | e1 = SELF; "lxor"; e2 = SELF -> <:expr< $e1$ lxor $e2$ >>
      | e1 = SELF; "mod"; e2 = SELF -> <:expr< $e1$ mod $e2$ >>
      | e1 = SELF; op = infixop3; e2 = SELF -> <:expr< $lid:op$ $e1$ $e2$ >> ]
    | "**" RIGHTA
      [ e1 = SELF; "**"; e2 = SELF -> <:expr< $e1$ ** $e2$ >>
      | e1 = SELF; "asr"; e2 = SELF -> <:expr< $e1$ asr $e2$ >>
      | e1 = SELF; "lsl"; e2 = SELF -> <:expr< $e1$ lsl $e2$ >>
      | e1 = SELF; "lsr"; e2 = SELF -> <:expr< $e1$ lsr $e2$ >>
      | e1 = SELF; op = infixop4; e2 = SELF -> <:expr< $lid:op$ $e1$ $e2$ >> ]
    | "unary minus" NONA
      [ "-"; e = SELF -> <:expr< $mkumin loc "-" e$ >>
      | "-."; e = SELF -> <:expr< $mkumin loc "-." e$ >> ]
    | "apply" LEFTA
      [ e1 = SELF; e2 = SELF -> (
          match constr_expr_arity e1 with
           1 -> <:expr< $e1$ $e2$ >>
          | _ ->
              ( match e2 with
               <:expr< ( $list:el$ ) >> ->
                  List.fold_left (fun e1 e2 -> <:expr< $e1$ $e2$ >>) e1 el
              | _ -> <:expr< $e1$ $e2$ >> ))
      | "assert"; e = SELF ->
          let f = <:expr< $str:!input_file$ >> in
          let bp = <:expr< $int:string_of_int (fst loc)$ >> in
          let ep = <:expr< $int:string_of_int (snd loc)$ >> in
          let raiser = <:expr< raise (Assert_failure ($f$, $bp$, $ep$)) >> in
          (match e with
           <:expr< False >> -> raiser
          | _ ->
              if !no_assert then <:expr< () >>
              else <:expr< if $e$ then () else $raiser$ >>) 
      | "lazy"; e = SELF ->
          <:expr< Pervasives.ref (Lazy.Delayed (fun () -> $e$)) >> ]
    | "." LEFTA
      [ e1 = SELF; "."; "("; e2 = SELF; ")" -> <:expr< $e1$ .( $e2$ ) >>
      | e1 = SELF; "."; "["; e2 = SELF; "]" -> <:expr< $e1$ .[ $e2$ ] >>
(**$      | e1 = SELF; "."; "{"; e2 = SELF; "}" -> bigarray_get loc e1 e2  ***)
      | e1 = SELF; "."; e2 = SELF -> <:expr< $e1$ . $e2$ >> ]
    | "~-" NONA
      [ "!"; e = SELF -> <:expr< $e$ . val>>
      | "~-"; e = SELF -> <:expr< ~- $e$ >>
      | "~-."; e = SELF -> <:expr< ~-. $e$ >>
      | f = prefixop; e = SELF -> <:expr< $lid:f$ $e$ >> ]
    | "simple" LEFTA
      [ s = INT -> <:expr< $int:s$ >>
      | s = FLOAT -> <:expr< $flo:s$ >>
      | s = STRING -> <:expr< $str:s$ >>
      | c = CHAR -> <:expr< $chr:c$ >>
      | UIDENT "True" -> <:expr< $uid:" True"$ >>
      | UIDENT "False" -> <:expr< $uid:" False"$ >>
      | i = expr_ident -> i
      | s = "false" -> <:expr< False >>
      | s = "true" -> <:expr< True >>
      | "["; "]" -> <:expr< [] >>
      | "["; el = sync_expr1_semi_list; "]" -> <:expr< $mklistexp loc None el$ >>
      | "[|"; "|]" -> <:expr< [| |] >>
      | "[|"; el = sync_expr1_semi_list; "|]" -> <:expr< [| $list:el$ |] >>
(***      | "{"; test_label_eq; lel = lbl_expr_list; "}" ->
          <:expr< { $list:lel$ } >>
      | "{"; e = expr LEVEL "."; "with"; lel = lbl_expr_list; "}" ->
          <:expr< { ($e$) with $list:lel$ } >>
***)      | "("; ")" -> <:expr< () >>
      | "("; op = operator_rparen -> <:expr< $lid:op$ >>
      | "("; e = SELF; ":"; t = ctyp; ")" -> <:expr< ($e$ : $t$) >>
      | "("; e = SELF; ")" -> <:expr< $e$ >>
      | "begin"; e = SELF; "end" -> <:expr< $e$ >>
      | x = LOCATE ->
          let x =
            try
              let i = String.index x ':' in
              (int_of_string (String.sub x 0 i),
               String.sub x (i + 1) (String.length x - i - 1))
            with
             Not_found | Failure _ -> (0, x) 
          in
          Pcaml.handle_expr_locate loc x
      | x = QUOTATION ->
          let x =
            try
              let i = String.index x ':' in
              (String.sub x 0 i,
               String.sub x (i + 1) (String.length x - i - 1))
            with
             Not_found -> ("", x) 
          in
          Pcaml.handle_expr_quotation loc x ] ]
;

  expr_ident:
    [ RIGHTA
      [ i = LIDENT -> <:expr< $lid:i$ >>
      | i = UIDENT -> <:expr< $uid:i$ >>
      | i = UIDENT; "."; j = SELF ->
          let rec loop m exp = match exp with 
             <:expr< $x$ . $y$ >> -> loop <:expr< $m$ . $x$ >> y
            | e -> <:expr< $m$ . $e$ >> 
          in
          loop <:expr< $uid:i$ >> j
      | i = UIDENT; "."; "("; j = operator_rparen ->
          <:expr< $uid:i$ . $lid:j$ >> ] ]
  ;

  sync_expr1_semi_list:
    [ [ e = sync_expr LEVEL "sync_expr1"; ";"; 
	      el = sync_expr1_semi_list -> (e :: el)
      | e = sync_expr LEVEL "sync_expr1"; ";" -> [e]
      | e = sync_expr LEVEL "sync_expr1" -> [e] ] ]
  ;    
  expr :
    [ [ "sync"; "("; c = sync_expr; ")" ->
        incr num_sync; 
	   num_get := 0;
	 add_expr loc <:expr< 
	   let ___fun_sync () = $c$ in 
	   (match (Flight.sync_type.(0), Flight.sync_type.(1)) with
	         [ (0 , 0 ) -> do {
		   Flight.begin_sync (___local_shift + $int:string_of_int !num_sync$); 
	           let ___res_sync = (try ___fun_sync ()  with
		                [ Flight.RemoteExn e -> raise e
		               | e -> raise e ]) in
                              Flight.end_sync();  ___res_sync}
	       | (___s,___g) -> 
		   if ((___s == (___local_shift + $int:string_of_int !num_sync$)) 
                     && (___g == 0)) 
		     || (___g == (___local_shift + $int:string_of_int !num_sync$))
		   then ___fun_sync ()
		   else raise Flight.NestedSync ] )>> ] ]
  ;

END;

